# Copyright (c) SenseTime. All Rights Reserved.

from __future__ import absolute_import 
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging 
import os
import json

import cv2
import numpy as np
import torch
import torch.nn.functional as F

from pysot.core.config import cfg
from pysot.utils.anchor import Anchors
from pysot.tracker.siamrpn_tracker import SiamRPNTracker

class SiamRPNLTTracker(SiamRPNTracker):
    def __init__(self, model):
        super(SiamRPNLTTracker, self).__init__(model)
        self.longterm_state = False
        self.toPrint = True
        self.missing = 0
        self.found = 61
        self.im_patch = None

    def track(self, img):
        """
        args:
            img(np.ndarray): BGR image
        return:
            bbox(list):[x, y, width, height]
        """
        self.id += 1
        prt_str = ''
        w_z = self.size[0] + cfg.TRACK.CONTEXT_AMOUNT * np.sum(self.size)
        h_z = self.size[1] + cfg.TRACK.CONTEXT_AMOUNT * np.sum(self.size)
        s_z = np.sqrt(w_z * h_z)
        scale_z = cfg.TRACK.EXEMPLAR_SIZE / s_z
        
        if self.longterm_state:
            instance_size = cfg.TRACK.LOST_INSTANCE_SIZE
            penalty_k = 0
        else:
            if self.found > 60:
                instance_size = cfg.TRACK.INSTANCE_SIZE
                penalty_k = cfg.TRACK.PENALTY_K
            else:
                instance_size = cfg.TRACK.INSTANCE_SIZE*2
                penalty_k = 0

        score_size = (instance_size - cfg.TRACK.EXEMPLAR_SIZE) // \
                cfg.ANCHOR.STRIDE + 1 + cfg.TRACK.BASE_SIZE
        hanning = np.hanning(score_size)
        window = np.outer(hanning, hanning)
        window = np.tile(window.flatten(), self.anchor_num)
        anchors = self.generate_anchor(score_size)

        s_x = s_z * (instance_size / cfg.TRACK.EXEMPLAR_SIZE)
        
        x_crop = self.get_subwindow(img, self.center_pos, instance_size,
                round(s_x), self.channel_average)
        outputs = self.model.track(x_crop)
        score = self._convert_score(outputs['cls'])
        pred_bbox = self._convert_bbox(outputs['loc'], anchors)

        def change(r):
            return np.maximum(r, 1. / r)

        def sz(w, h):
            pad = (w + h) * 0.5
            return np.sqrt((w + pad) * (h + pad))
        
        # scale penalty
        s_c = change(sz(pred_bbox[2,:], pred_bbox[3,:]) / 
                (sz(self.size[0]*scale_z, self.size[1]*scale_z))) 
        # ratio penalty
        r_c = change((self.size[0]/self.size[1]) /
                     (pred_bbox[2,:]/pred_bbox[3,:]))
        penalty = np.exp(-(r_c * s_c - 1) * penalty_k)
        pscore = penalty * score

        # window 
        if not self.longterm_state and self.found > 60:
            pscore = pscore * (1 - cfg.TRACK.WINDOW_INFLUENCE) + \
                    window * cfg.TRACK.WINDOW_INFLUENCE
        else:
            pscore = pscore * (1 - 0.000) + window * 0.000
        best_idx = np.argmax(pscore)

        bbox = pred_bbox[:, best_idx] / scale_z
        lr = penalty[best_idx] * score[best_idx] * cfg.TRACK.LR
       
        best_score = score[best_idx]
        if best_score >= cfg.TRACK.CONFIDENCE_LOW:
            cx = bbox[0] + self.center_pos[0]
            cy = bbox[1] + self.center_pos[1]

            width = self.size[0] * (1 - lr) + bbox[2] * lr
            height = self.size[1] * (1 - lr) + bbox[3] * lr
        else:
            cx = self.center_pos[0]
            cy = self.center_pos[1]

            width = self.size[0]
            height = self.size[1]

        self.center_pos = np.array([cx, cy])
        self.size = np.array([width, height])
        
        cx, cy, width, height = self._bbox_clip(cx, cy, width, height, img.shape[:2])
        bbox = [cx - width / 2,
                cy - height / 2,
                width,
                height]

        if best_score < cfg.TRACK.CONFIDENCE_LOW:
            self.longterm_state = True
        elif best_score > cfg.TRACK.CONFIDENCE_HIGH or (self.missing>400 and best_score > 0.9):
            self.longterm_state = False
        
        if self.longterm_state:
            self.missing += 1
            self.found = 0
            
            if self.missing < 40:
                self.size = self.pre_size
                self.center_pos = self.pre_center_pos
            elif self.missing > 90:
                o_w = self.min_size[0]
                o_h = self.min_size[1]
                r_w = self.max_size[0]
                r_h = self.max_size[1]
                rand_ratio = np.random.rand()
                self.size  = [rand_ratio*np.abs(r_w-o_w)+min(r_w, o_w), rand_ratio*np.abs(r_h-o_h)+min(r_h, o_h)]
                
                x,y = self.pre_center_pos
                w,h = self.size
                mul = int(self.missing/60)
                self.center_pos = [x + (np.random.rand()-0.5)*w*mul, y + (np.random.rand()-0.5)*h*mul]
                if self.missing > 150:
                
                    dis = np.array([np.linalg.norm(self.pre_center_pos-np.array(pt)) for pt in self.five_focus_pt])
                    x, y = self.five_focus_pt[np.argmin(dis)]
                    w, h = self.size
                    mul = int(self.missing/60)
                    self.center_pos = [x + (np.random.rand()-0.5)*w*mul, y + (np.random.rand()-0.5)*h*mul]
        else:
            self.missing = 0
            self.found += 1
            ### To avoid tracking the fix label in the frame ###
            if self.id % 30 == 0:
                if len(self.center_error)>=4:
                    self.center_error.pop(0)
                self.center_error.append(np.linalg.norm(self.pre_center_pos-self.center_pos))
                
                if len(self.center_error)>=4 and np.average(self.center_error)<0.25:
                    
                    h, w = img.shape[:2]
                    w_pad = w/5
                    h_pad = h/5
                    x, y  = self.center_pos
                    if x < w_pad or x > w-w_pad or y < h_pad or y > h-h_pad:
                        self.longterm_state = True
            if self.toPrint and len(self.center_error)>=4: prt_str += 'center_error: '+str(np.average(self.center_error))+'\n'
            ### To avoid tracking the fix label in the frame ###
            
            self.pre_size = self.size
            self.pre_center_pos = self.center_pos
            h, w = img.shape[:2]
            w_pad = w/20
            h_pad = h/20
            x, y  = self.center_pos
            if x > w_pad and x < w-w_pad and y > h_pad and y < h-h_pad:
                if np.sum(self.min_size) > np.sum(self.size):
                    self.min_size = self.size
                if np.sum(self.max_size) < np.sum(self.size):
                    self.max_size = self.size
                    
        if self.toPrint: prt_str += 'found: '+str(self.found)+'\n'
        if self.toPrint: prt_str += 'missing: '+str(self.missing)+'\n'
        if self.toPrint: prt_str += 'score: '+str(best_score)+'\n'
        if self.toPrint: prt_str += str(self.longterm_state)+'\n'
        
        if self.longterm_state: best_score = 0
        
        return {
                'bbox': bbox,
                'best_score': best_score,
                'prt_str': prt_str,
                'im_patch': self.im_patch
               }
